
import sys

import biolqm

from colomoto_jupyter import *
from colomoto_jupyter.sessionfiles import new_output_file
from colomoto_jupyter.io import ensure_localfile

from colomoto import ModelState

from .jupyter import upload
from .gateway import japi, restart

import pandas as pd

import ginsim.state

from py4j.java_gateway import JavaObject

def load(filename, *args):
    filename = ensure_localfile(filename)
    obj = japi.gs.load(filename, *args)
    assert obj is not None, "Error while loading model"
    return obj

def service(name):
    return japi.gs.service(name)

def get_image(lrg, state=None, style=None):
    if style is not None:
        return japi.gs.service("image").rawPNG(lrg, style)

    if state is not None:
        if isinstance(state, dict):
            state = ginsim.state.get_ginsim_state(lrg, state)
        elif isinstance(state, pd.Series):
            # FIXME: here we assume that the Series is using the proper order
            # Fixing the index could be done with:
            # state = state.reindex( ["Proper", "Node", "Order"], fill_value=-1 )
            state = state.values.tobytes()
        return japi.gs.service("image").rawPNG(lrg, state)

    return japi.gs.service("image").rawPNG(lrg)

def lrg_style(lrg):
    return japi.gs.service("image").customStyleProvider(lrg)

def is_ginsim_object(obj):
    """
    Returns true if argument is a GINsim object
    """
    return isinstance(obj, JavaObject) \
        and obj.getClass().getPackage().getName() == "org.ginsim.core.graph.regulatorygraph"

if IN_IPYTHON:
    def show(lrg, state=None, style=None):
        return show_image(get_image(lrg, state, style))

def to_biolqm(lrg):
    return lrg.getModel()

def to_maboss(lrg):
    return biolqm.to_maboss(lrg.getModel())

def to_pint(lrg, simplify=True):
    return biolqm.to_pint(lrg.getModel(), simplify)

def to_minibn(lrg, **opts):
    return biolqm.to_minibn(lrg.getModel(), **opts)

__nusmvReserved = [ "MODULE", "DEFINE", "MDEFINE",
    "CONSTANTS", "VAR", "IVAR", "FROZENVAR", "INIT", "TRANS", "INVAR",
    "SPEC", "CTLSPEC", "LTLSPEC", "PSLSPEC", "COMPUTE", "NAME",
    "INVARSPEC", "FAIRNESS", "JUSTICE", "COMPASSION", "ISA", "ASSIGN",
    "CONSTRAINT", "SIMPWFF", "CTLWFF", "LTLWFF", "PSLWFF", "COMPWFF",
    "IN", "MIN", "MAX", "MIRROR", "PRED", "PREDICATES", "process",
    "array", "of", "boolean", "integer", "real", "word", "word1",
    "bool", "signed", "unsigned", "extend", "resize", "sizeof",
    "uwconst", "swconst", "EX", "AX", "EF", "AF", "EG", "AG", "E", "F",
    "O", "G", "H", "X", "Y", "Z", "A", "U", "S", "V", "T", "BU", "EBF",
    "ABF", "EBG", "ABG", "case", "esac", "mod", "next", "init",
    "union", "in", "xor", "xnor", "self", "TRUE", "FALSE", "count"]
__nusmvReserved = [k.lower() for k in __nusmvReserved]

def nusmv_var_state(ai):
    a, i = ai
    if len(a) == 1 or a.lower() in __nusmvReserved:
        a = "_%s" % a
    return "%s=%s" % (a,i)

def to_nusmv(lrg, update_mode="async"):
    assert update_mode in ["async", "sync", "complete"], "Unknown update mode"

    serv = service("NuSMV")
    cfg = serv.getConfig(lrg.getModel())

    ## configure update mode
    d_update = {
        "async": cfg.CFG_ASYNC,
        "sync": cfg.CFG_SYNC,
        "complete": cfg.CFG_COMPLETE,
    }
    cfg.setUpdatePolicy(d_update[update_mode])

    ## configure input nodes
    sinit = japi.gs.associated(lrg, "initialState", True)
    for inputNode in sinit.getInputNodes():
        cfg.addFixedInput(str(inputNode))

    smvfile = new_output_file("smv")
    serv.run(cfg, smvfile)
    from colomoto.modelchecking import ColomotoNuSMV
    return ColomotoNuSMV(smvfile, nusmv_var_state)




from colomoto import ModelState

import ginsim
import py4j

import sys
import biolqm

from ginsim.gateway import register


def get_ginsim_state(lrg, state):
    order = lrg.getNodeOrder()
    l = len(order)
    gw = ginsim.japi.java
    int_class = gw.jvm.int
    int_array = gw.new_array(int_class, l)
    i = 0
    for node in order:
        uid = node.getId()
        if uid in state:
            int_array[i] = state[uid]
        i += 1
    
    return int_array

def get_model_state(nodes, state):
    mstate = ModelState()
    l = len(nodes)
    for i in range(l):
        mstate[ nodes[i].getNodeID() ] = state[i]
    
    return mstate





from colomoto_jupyter import IN_IPYTHON, jupyter_setup

if IN_IPYTHON:
    menu = [
        {"name": "Upload model",
            "snippet": ["lrg = ginsim.upload()"]},
        {"name": "Load model",
            "snippet": ["lrg = ginsim.load(\"model.zginml\")"]},
        "---",
        {"name":"Export to file",
            "sub-menu": [
            {"name": "MaBoSS (.bnd, .cfg)",
                "snippet":['ginsim.service("maboss").export(lrg, "mymodel.bnd")']},
            ]},
        {"name":"Convert to tool",
            "sub-menu": [
            {"name": "bioLQM",
                "snippet":['lqm = ginsim.to_biolqm(lrg)']},
            {"name": "MaBoSS",
                "snippet":['masim = ginsim.to_maboss(lrg)']},
            {"name": "NuSMV",
                "snippet":['smv = ginsim.to_nusmv(lrg)']},
            {"name": "Pint",
                "snippet":['an = ginsim.to_pint(lrg)']},
            ]},
        "---",
        {"name": " Display regulatory graph",
            "snippet": ["ginsim.show(lrg)"]},
    ]
    toolbar = [
        {"name": "upload", "setup": {
            "icon": "fa-upload",
            "help": "Upload model",
            "handler": "action_upload_model"}},
    ]
    js_api = {
    "action_upload_model": """function() {
        var cell = Jupyter.notebook.get_selected_cell();
        cell.set_text('lrg = '+ginsim_jsapi.module_alias+'.upload()');
        cell.focus_editor();
    }""",
    }
    jupyter_setup("ginsim",
        label="GINsim",
        color="blue", # for menu and toolbar
        menu=menu,
        toolbar=toolbar,
        js_api=js_api)


    from colomoto_jupyter.upload import jupyter_upload
    def upload():
        return jupyter_upload("upload", "load")

else:
    def upload():
        raise NotImplementedError





import sys

from colomoto_jupyter.sessionfiles import new_output_file
from colomoto_jupyter.io import ensure_localfile
from colomoto_jupyter import import_colomoto_tool

from .jupyter import upload

import pandas as pd
import numpy as np

from ginsim.gateway import japi, restart
from ginsim.state import *

from py4j.java_gateway import JavaObject

class LQMTool:
    def __init__(self, tool):
        self._uid = tool.getID()
        self._tool = tool

    def getTask(self, model, parameters=None):
        if parameters:
            task = self._tool.getTask(model, parameters)
        else:
            task = self._tool.getTask(model)
        
        if self._uid in _japi_converters:
            return LQMTask(task, _japi_converters[self._uid])
        
        return LQMTask(task)

    def __call__(self, model, parameters=None, autoconvert=True):
        return self.getTask(model, parameters)(autoconvert)

    def __getattr__(self, name):
        return self._tool.__getattr__(name)

class LQMTask:
    def __init__(self, task, convert=None):
        self._task = task
        self._convert = convert

    def __call__(self, autoconvert=True):
        result = self._task.call()
        
        if autoconvert and self._convert:
            return self._convert(result)
        
        return result

    def __getattr__(self, name):
        return self._task.__getattr__(name)

class LQMModifier:
    def __init__(self, modifier):
        self._modifier = modifier

    def getModifier(self, model, parameters=None):
        if parameters:
            mod = self._modifier.getModifier(model, parameters)
        else:
            mod = self._modifier.getModifier(model)
        return mod

    def __call__(self, model, parameters=None):
        return self.getModifier(model, parameters).call()
    
    def __getattr__(self, name):
        return self._modifier.__getattr__(name)


def states_to_dataframe(states):
    if states == None or len(states) < 1:
        return []
    
    ids = [str(n) for n in states.getComponents()]
    data = [ np.frombuffer(states.fillState(None,idx), dtype=np.byte) for idx in range(len(states)) ]
    data = np.concatenate(data)
    data = data.reshape( (states.size(), len(ids)) )
    ids = [n.getNodeID() for n in states.getComponents()]
    return pd.DataFrame(data, columns=ids)


def convert_fixpoints(stables):
    if stables == None or len(stables) < 1:
        return []
    return [ get_model_state(stables.getComponents(), state) for state in stables ]

def convert_trapspace(traps):
    if traps == None or len(traps) < 1:
        return []
    return [ get_model_state(traps.nodes, state.pattern) for state in traps ]


class SimulationIterable:
    def __init__(self, simulation):
        self.simulation = simulation
    
    def __iter__(self):
        return SimulationIterator(self.simulation)

class SimulationIterator:
    def __init__(self, simulation):
        self.iterator = simulation.iterator()
        self.nodes = simulation.getNodes()
    
    def __next__(self):
        if self.iterator.hasNext():
            return get_model_state(self.nodes, self.iterator.next())
        raise StopIteration()

def convert_trace(simulation):
    if simulation == None:
        return []
    return SimulationIterable(simulation)

_japi_wrappers = set()
_japi_converters = {
    "fixpoints": convert_fixpoints,
    "trapspaces": convert_trapspace,
    "trace": convert_trace,
    "random": convert_trace,
}

def _japi_start():
    current_module = __import__(__name__)
    
    for tool in japi.java.jvm.org.colomoto.biolqm.service.LQMServiceManager.getTools():
        name = tool.getID()
        wrapper = LQMTool(tool)
        setattr(current_module, name, wrapper)
        _japi_wrappers.add(name)
        aliases = tool.getAliases()
        if aliases:
            for alias in aliases:
                setattr(current_module, alias, wrapper)

    for mod in japi.java.jvm.org.colomoto.biolqm.service.LQMServiceManager.getModifiers():
        name = mod.getID()
        wrapper = LQMModifier(mod)
        setattr(current_module, name, wrapper)
        _japi_wrappers.add(name)

def _japi_stop():
    current_module = __import__(__name__)
    for name in _japi_wrappers:
        delattr(current_module, name)
    _japi_wrappers.clear()

def load(filename, *args):
    filename = ensure_localfile(filename)
    obj = japi.lqm.load(filename, *args)
    assert obj is not None, "Error while loading model"
    return obj

def save(model, filename, format=None):
    assert japi.lqm.save(model, filename, format)
    return filename

def modify(model, modifier, *args):
    return japi.lqm.modify(model, modifier, *args)

def getFormat(name):
    return japi.lqm.getFormat(name)

def getModifier(name):
    return japi.lqm.getModifier(name)

def getTool(name):
    return japi.lqm.getTool(name)

def is_biolqm_object(obj):
    """
    Returns true if argument is a bioLQM object
    """
    return isinstance(obj, JavaObject) \
        and obj.getClass().getPackage().getName() == "org.colomoto.biolqm"

def influence_graph(model):
    import networkx
    matrix = japi.java.jvm.org.colomoto.biolqm.ConnectivityMatrix(model)
    cnodes = model.getComponents()
    enodes = model.getExtraComponents()
    edges = []
    for (nodes, extra) in [(cnodes, False), (enodes, True)]:
        for idx in range(len(nodes)):
            a  = nodes.get(idx).getNodeID()
            regulators = matrix.getRegulators(idx, extra)
            if not regulators:
                continue
            effects = matrix.getRegulatorEffects(idx, extra)
            for i in range(len(regulators)):
                reg = regulators[i]
                b = nodes.get(reg).getNodeID()
                curEffects = effects[i]
                for v in range(len(curEffects)):
                    e = curEffects[v]
                    if e == e.NONE:
                        continue
                    elif e == e.POSITIVE:
                        sign = 1
                    elif e == e.NEGATIVE:
                        sign = -1
                    else:
                        sign = 0
                    edges.append((b,sign,a))
    ig = networkx.MultiDiGraph()
    labels = {1: "+", -1: "-", 0: "?"}
    for (b, sign, a) in edges:
        ig.add_edge(b, a, sign=sign, label=labels[sign])
    return ig

default_layout = "neato"

def autolayout(model, method=default_layout, scale="auto"):
    from colomoto.helpers import layout_graph
    if scale == "auto":
        default = 1
        scales = {
            "dot": 0.25,
            "circo": 0.4,
            "neato": 0.55,
        }
        scale = scales.get(method, default)
    ig = influence_graph(model)
    layout = layout_graph(ig, method=method, scale=scale*200)
    add_layout(model, layout)

def to_ginsim(model, ensure_layout=True,
        layout_method=default_layout, layout_scale="auto"):
    """
    Convert a bioLQM model into an equivalent GINsim model using the
    :py:mod:`ginsim` Python module.
    """
    if ensure_layout and not model.hasLayout():
        autolayout(model, method=layout_method, scale=layout_scale)
    return japi.gs.load(model)

def to_maboss(model):
    maboss = import_colomoto_tool("maboss")
    maboss_file = new_output_file("bnd")
    assert save(model, maboss_file, "bnd")
    return maboss.load(maboss_file, "%s.cfg" % maboss_file)

def to_pint(model, simplify=True):
    anfile = new_output_file("an")
    assert save(model, anfile, "an")
    pypint = import_colomoto_tool("pypint")
    an = pypint.load(anfile)
    if simplify:
        an = an.simplify()
    return an

def to_minibn(model, simplify=True, ensure_boolean=False):
    from colomoto import minibn
    if ensure_boolean or model.isBoolean():
        fmt = "bnet"
        cls = minibn.BooleanNetwork
    else:
        fmt = "mnet"
        cls = minibn.MultiValuedNetwork
    bnfile = new_output_file(fmt)
    assert save(model, bnfile, fmt)
    with open(bnfile) as data:
        bn = cls(data)
    if simplify and hasattr(bn, "simplify"):
        bn = bn.simplify()
    return bn


def add_layout(model, layout):
    """
    Add layout information to a model
    The layout object should be a dict of dicts in which the 'nodes' entry is
    a dict giving the (x, y) position of all nodes, identified by their uid.
    """
    inner = model.getLayout()
    for uid, values in layout['nodes'].items():
        ni = model.getComponent(uid)
        if ni:
            inner.setPosition(ni, int(values['x']), int(values['y']))
        else:
            print("no such node: ", uid)


from ginsim.gateway import register
register(sys.modules[__name__])
del(register)





from colomoto_jupyter import IN_IPYTHON, jupyter_setup

if IN_IPYTHON:

    def export_entry(d, f, e=None):
        if e is None:
            e = f
        return {"name": "{} (.{})".format(d, e),
            "snippet": ['biolqm.save(lqm, "mymodel.{}", "{}")'.format(e, f)]}

    menu = [
        {"name": "Upload model",
            "snippet": ["lqm = biolqm.upload()"]},
        {"name": "Load model",
            "snippet": ["lqm = biolqm.load(\"model.sbml\")"]},
        "---",
        {"name":"Export to file",
            "sub-menu": [
                export_entry("SBML-qual v1.0", "sbml"),
                "Functions formats",
                export_entry("BoolNet", "bnet"),
                export_entry("BooleanNet", "booleannet"),
                export_entry("BoolSim", "boolsim"),
                export_entry("Raw Boolean functions", "boolfunctions"),
                export_entry("Truth table", "tt"),
                "Petri net formats",
                export_entry("APNN", "apnn"),
                export_entry("INA", "ina"),
                export_entry("PNML", "pnml"),
                "Dedicated formats",
                export_entry("GINML", "ginml"),
                export_entry("GNA non-xml", "gna"),
                export_entry("MaBoSS", "bnd"),
                export_entry("Pint", "an"),
            ]},
        {"name":"Convert to tool",
            "sub-menu": [
            {"name": "GINsim", "snippet": [
                'lrg = biolqm.to_ginsim(lqm)']},
            {"name": "MaBoSS", "snippet": [
                'masim = biolqm.to_maboss(lqm)']},
            {"name": "Pint", "snippet": [
                'an = biolqm.to_pint(lqm)']},
            ]},
    "---",
    {"name": "Compute fixpoints",
    #{"name": "Compute fixpoints (MDD method)",
        "snippet": ["fps = biolqm.fixpoints(lqm)"]},
    #{"name": "Compute fixpoints (ASP method)",
    #    "snippet": ["fps = biolqm.fixpoints(lqm, 'asp')"]},
    {"name": "Compute trap spaces",
        "snippet": ["traps = biolqm.trapspace(lqm)"]},
    "---",
    {"name":"Model modifications",
        "sub-menu": [
            {"name": "Perturbation",
                "snippet": ["lqm_mod = biolqm.perturbation(lqm, \"node%0\")"]},
            {"name": "Booleanization",
                "snippet": ["lqm_bool = biolqm.booleanize(lqm)"]},
            {"name": "Reduction",
                "snippet": ["lqm_red = biolqm.reduce(lqm, \"fixed,output,duplicate\")"]},
            {"name": "Reversal",
                "snippet": ["lqm_rev = biolqm.reverse(lqm)"]},
            {"name": "Sanitize",
                "snippet": ["lqm_san = biolqm.sanitize(lqm)"]},
        ]},
     "---",
     {"name": "Documentation",
        "external-link": "http://ginsim.naldi.info/biolqm/site/doc/index.html"}
    ]
    toolbar = [
        {"name": "upload", "setup": {
            "icon": "fa-upload",
            "help": "Upload model",
            "handler": "action_upload_model"}},
    ]
    js_api = {
    "action_upload_model": """function() {
        var cell = Jupyter.notebook.get_selected_cell();
        cell.set_text('lqm = '+biolqm_jsapi.module_alias+'.upload()');
        cell.focus_editor();
    }""",
    }
    jupyter_setup("biolqm",
        label="bioLQM",
        color="#00007f", # for menu and toolbar
        menu=menu,
        toolbar=toolbar,
        js_api=js_api)


    from colomoto_jupyter.upload import jupyter_upload
    def upload():
        return jupyter_upload("upload", "load")

else:
    def upload():
        raise NotImplementedError





from setuptools import setup, find_packages

NAME = 'ginsim'

setup(name=NAME,
    version='0.4.1',
    author = "Aurélien Naldi",
    author_email = "aurelien.naldi@gmail.com",
    url = "https://github.com/GINsim/GINsim-python",
    description = "Python interface to GINsim and BioLQM API",
    long_description = """
Provides interface to Java API of GINsim and BioLQM
""",
    install_requires = [
        "py4j",
        "colomoto_jupyter",
    ],
    packages = find_packages(),
    classifiers=[
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    keywords="computational systems biology",
)





import atexit
import subprocess
from subprocess import PIPE

from py4j.java_gateway import JavaGateway, GatewayParameters

__env = {}
__registered = []

class LQMProxy(object):
    java = None
    gs = None
    lqm = None
    def clear(self):
        self.java = None
        self.gs = None
        self.lqm = None

japi = LQMProxy()

def start():
    assert (not __env)
    __env["proc"] = subprocess.Popen(["GINsim", "-py"], \
                        stdout=PIPE, stdin=PIPE, stderr=PIPE)
    port = int(__env["proc"].stdout.readline().strip())

    # start the gateway and return the entry point (GINsim's ScriptLauncher)
    param = GatewayParameters(port=port, auto_convert=True, auto_field=True)
    __env["gw"] = JavaGateway(gateway_parameters=param)

    global japi
    japi.java = __env["gw"]
    japi.gs = __env["gw"].entry_point
    japi.lqm = japi.gs.LQM()
    for module in __registered:
        module._japi_start()

def stop():
    if not __env:
        return
    global japi
    japi.clear()
    __env["gw"].shutdown()
    __env["proc"].terminate()
    __env.clear()
    for module in __registered:
        module._japi_stop()

def restart():
    stop()
    start()

def register(module):
    __registered.append(module)
    if __env:
        module._japi_start()

start()
atexit.register(stop)

